#!/usr/bin/perl
use warnings;
use strict;
use lib '../include';
use Net::Telnet::Cisco;
use Net::Ping;
use Term::ANSIColor;
use nms qw(switch_connect switch_exec switch_disconnect);

BEGIN {
	require "../include/config.pm";
	eval {
		require "../include/config.local.pm";
	};
}

my $search_path = ".infra.$nms::config::tgname.gathering.org";

# Dlink DGS-3100 default values
my $default_ip = '10.90.90.90';
my $router_dlink1g_ip = '10.90.90.91';
my $dlink1g_netmask = '255.255.255.0';

my $patchlist = $ARGV[0] or die "First argument must be patchlist file";
my $single_switch = $ARGV[1] || undef;

$| = 1;

sub ios_ping {
	my ($ios, $ip, $timeout) = @_;
	my $pong = 0;
	my $tries = 0;
	print "Pinging $ip";
	while ($pong == 0 && $tries < $timeout) {
		print ".";
		my @res = $ios->cmd("ping $ip repeat 1 timeout 1");
		$pong = 1 if ($res[-1] =~ /Success rate is 100 percent/);
		$tries++;
	}
	print "\n";
	return $pong;
}

sub dlink_cmd {
	my ($dlink, $cmd) = @_;
	
	print "Running command on dlink: $cmd\n";
	if (!$dlink->cmd($cmd)) {
		print "Dlink command $cmd failed: ".$dlink->errmsg;
		return 0;
	}
	return 1;
}

sub setup {
	my %switch = @_;
	print "\n";
	print "Configuring $switch{switchname}\n";
	print "Pinging $switch{ipv4address}\n";
	my $p = Net::Ping->new();
	if ($p->ping($switch{ipv4address}, 5)) {
		print colored('INFO', 'bold green'), ": Skipping $switch{switchname} IP '$switch{ipv4address}' is already responding to ping! \\o/\n";
		return 1;
	}
	print "Not responding, configuring device\n";

	print "Connection to IOS on $switch{coreswip}\n";
	my $ios = Net::Telnet::Cisco->new(Host => $switch{coreswip}.$search_path,
			Errmode => 'return',
			output_log => "output-$switch{coreswip}.log",
			input_log => "input-$switch{coreswip}.log",
			Prompt => '/\S+[#>]/');
	if (!defined($ios)) {
		print STDERR colored('ERROR', 'red'), "/", colored($switch{switchname}, 'bold'), ": Could not connect to $switch{coreswip}\n";
		return 0;
	}

	print "Logging in to IOS\n";
	$ios->login($nms::config::ios_user, $nms::config::ios_pass) or die "Can't log in to $switch{coreswip}";
	$ios->enable;
	print $ios->errmsg;		

	# Disable paging
	$ios->cmd("terminal length 0");

	print "Setting up ports\n";
	# Turn off port 2
	$ios->cmd("conf t");
	$ios->cmd("default int $switch{port1}");
	$ios->cmd("default int $switch{port2}");
	$ios->cmd("no int Po$switch{etherchannel}");
	$ios->cmd("int $switch{port2}");
	$ios->cmd("shut");
	$ios->cmd("no switchport");
	print $ios->errmsg;
	# Set up port 1 for Dlink telneting
	$ios->cmd("int $switch{port1}");
	$ios->cmd("no switchport");
	$ios->cmd("ip address $router_dlink1g_ip 255.255.255.0");
	$ios->cmd("no shut");
	$ios->cmd("end");
	print($ios->errmsg);

	print "Waiting for Dlink on $default_ip\n";
	if (!ios_ping($ios, $default_ip, 30)) {
		# Take down interface
		print "Couldn't connect to $switch{switchname} on $default_ip, resetting interfaces $switch{port1} and $switch{port2} and aborting\n";
		$ios->cmd("conf t");
		$ios->cmd("default int $switch{port1}");
		$ios->cmd("int $switch{port1}");
		$ios->cmd("shut");
		$ios->cmd("default int $switch{port2}");
		$ios->cmd("int $switch{port2}");
		$ios->cmd("shut");
		$ios->cmd("end");
		return 0;
	}

	# Telnet to Dlink
	print("Starting D-Link config phase 1\n");
	my $dlink = telnet_login(\%switch, $default_ip) or return 0;
	$dlink->print("disable clipaging");
	dlink_cmd($dlink, "create link_aggregation group_id 1 type lacp") or return 0;
	dlink_cmd($dlink, "config link_aggregation group_id 1 ports 1:(47-48)") or return 0;
	dlink_cmd($dlink, "enable stp") or return 0;
	print "STP enabled, logging in again\n";
	$dlink->close;

	if (!ios_ping($ios, $default_ip, 60)) {
		print "Can't login to $switch{switchname} on $default_ip, aborting\n";
		$ios->close;
		return 0;
	}

	$dlink = telnet_login(\%switch, $default_ip) or return 0;
	print "Running STP config\n";
	dlink_cmd($dlink, "config stp version mstp") or return 0;
	dlink_cmd($dlink, "config stp priority 0 instance_id 0") or return 0;
	dlink_cmd($dlink, "config stp ports 1:(1-44) edge true") or return 0;
	dlink_cmd($dlink, "enable lldp") or return 0;
	print "Setting IP address\n";
	$dlink->print("config ipif System ipaddress $switch{ipv4address}/30 vlan default");
	print "Closing Dlink link\n";
	$dlink->close;
	print($dlink->errmsg);

	print "Configuring IP on gateway\n";
	$ios->cmd("conf t");
	$ios->cmd("int $switch{port1}");
	$ios->cmd("ip address $switch{ipv4gateway} 255.255.255.128");
	$ios->cmd("end");

	# Wait for network convergence
	if (!ios_ping($ios, $switch{ipv4address}, 60)) {
		print "Can't ping $switch{switchname} on $switch{ipv4address}, aborting\n";
		$ios->close;
		return 0;
	}

	# Do it again!
	print("D-Link confing phase 2...\n");
	$dlink = telnet_login(\%switch, $switch{ipv4address}) or return 0;
	$dlink->print("create iproute default $switch{ipv4gateway}");
	$dlink->print("save");
	dlink_cmd($dlink, "Y");
	$dlink->print("logout");
	print $dlink->errmsg;

	print "Final IOS config phase\n";
	$ios->cmd("conf t");
	$ios->cmd("default interface $switch{port1}"); # Reset interface
	$ios->cmd("int $switch{port1}");
	$ios->cmd("no switchport");
	$ios->cmd("no shut");
	$ios->cmd("desc Dlink $switch{switchname}");
	$ios->cmd("channel-group $switch{etherchannel} mode passive");
	$ios->cmd("int $switch{port2}");
	$ios->cmd("no switchport");
	$ios->cmd("no shut");
	$ios->cmd("desc Dlink $switch{switchname}");
	$ios->cmd("channel-group $switch{etherchannel} mode passive");
	$ios->cmd("int po$switch{etherchannel}");
	$ios->cmd("desc Dlink $switch{switchname}");
	$ios->cmd("ip address $switch{ipv4gateway} 255.255.255.128");
	$ios->cmd("ipv6 enable");
	$ios->cmd("ipv6 address $switch{ipv6address}");
	$ios->cmd("ip helper-address $nms::config::dhcp_server1");
	$ios->cmd("no shut");
	$ios->cmd("end");
	$ios->close;

	print colored("SUCCESS:", 'green')." $switch{switchname} set up.\n";

	return 1;
}


sub telnet_login {
	my ($switch, $ip) = @_;

	my $dlink = Net::Telnet::Cisco->new(Host => $switch->{coreswip}.$search_path,
			Errmode => 'return',
			output_log => "output_telnet-".$switch->{coreswip}.".log",
			input_log => "input_telnet-".$switch->{coreswip}.".log",
			Prompt => '/\S+[#>]/',
			errmode => 'return',
			Timeout => 60);
	if (!defined($dlink)) {
		print STDERR colored('ERROR', 'red'), "/", colored($switch->{switchname}, 'bold'), ": Could not connect to $switch->{coreswip}\n";
		return 0;
	}

	$dlink->login($nms::config::ios_user, $nms::config::ios_pass) or die "Can't log in to $switch->{coreswip}";
	$dlink->enable;
	print "Telnetting..\n";
	$dlink->print("telnet $ip /source-interface $switch->{port1}");
	print "Waiting for login prompt\n";
	print $dlink->errmsg;
	$dlink->waitfor('/User ?Name:/') or return 0;
	print "Got login prompt, logging in\n";
	print $dlink->errmsg or return 0;
	$dlink->print("admin");
	$dlink->waitfor('/DGS-3100\#/') or return 0;
	print $dlink->errmsg;
	print "Logged in to Dlink\n";
	return $dlink;
}

my $only = shift;
print colored('INFO', 'bold blue'), ": IP run\n";
open (PATCHLIST, $patchlist) or die "Unable to open patchlist";
while (<PATCHLIST>) {
	next if /^ *#/; # skip comments
	my ($switchname, $coreswip, $port1, $port2, $etherchannel, $ipv4address, $ipv4gateway, $ipv6address) = split;
	my %switch = (
		switchname => $switchname,
		coreswip => $coreswip,
		port1 => $port1,
		port2 => $port2,
		etherchannel => $etherchannel,
		ipv4address => $ipv4address,
		ipv4gateway => $ipv4gateway,
		ipv6address => $ipv6address,
	);

	# Only configure a single switch?
	if (defined $single_switch) {
		if (!($single_switch eq $switchname)) {
			next;	
		}
	}

	if (!setup(%switch)) {
		print colored("FAILED:", "red")." Configuring $switchname failed.\n";
	}
}
close(PATCHLIST);
