//g++ -O2 -o planning planning.cpp && ./planning 13 34

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <limits.h>
#include <vector>
#include <map>
#include <algorithm>
#include <string>
#include <utility>

#define NUM_DISTRO 2
#define SWITCHES_PER_ROW 4
#define TRUNCATE_METRIC 1
#define TENPLUSTEN 250
#define THIRTYPLUSTEN 450
#define FIFTYPLUSTEN 650
#define FIFTYPLUSTHIRTY 850
#define FIFTYPLUSFIFTY 1050
#define INFINITY 99999

struct sw {
	unsigned char row, num;

	sw(unsigned char row, unsigned char num) : row(row), num(num) {}
};
int distro_placements[NUM_DISTRO] = {
	10, 26
};
unsigned horiz_cost[SWITCHES_PER_ROW] = {
	216, 72, 72+40, 216+40
};

// memoization table
struct key {
	unsigned char sw_ind, distro_bound, distro_bound_hard;
	unsigned char ports_left[NUM_DISTRO];

	bool operator< (const key &other) const
	{
		if (sw_ind != other.sw_ind)
			return (sw_ind < other.sw_ind);
		if (distro_bound != other.distro_bound)
			return (distro_bound < other.distro_bound);
		if (distro_bound_hard != other.distro_bound_hard)
			return (distro_bound_hard < other.distro_bound_hard);
		for (unsigned i = distro_bound_hard; i < NUM_DISTRO - 1; ++i)
			if (ports_left[i] != other.ports_left[i])
				return (ports_left[i] < other.ports_left[i]);
		return (ports_left[NUM_DISTRO - 1] < other.ports_left[NUM_DISTRO - 1]);
	}
};
struct value {
	unsigned char distro;
	unsigned cost, distance, this_cost, this_distance;
};
std::map<key, value> cache;
unsigned cache_hits = 0;
std::vector<sw> switches;
std::map<unsigned, unsigned> num_ports_used;

unsigned find_distance(sw from_where, unsigned distro)
{
	// 3.7m from row to row (2.5m gap + 1.2m boards)
	unsigned base_cost = 37 * abs(from_where.row - distro_placements[distro]) + horiz_cost[from_where.num];
	
	// 4m, 5m, 4m gaps (1.5m, 2.5m, 1.5m extra)
	if ((from_where.row <= 5) == (distro_placements[distro] >= 6))
		base_cost += 15;
	if ((from_where.row <= 13) == (distro_placements[distro] >= 14))
		base_cost += 15;
	if ((from_where.row <= 21) == (distro_placements[distro] >= 22))
		base_cost += 25;
	if ((from_where.row <= 29) == (distro_placements[distro] >= 30))
		base_cost += 15;

	// add 5m slack.
	return base_cost + 50;
}

unsigned cable_select(unsigned distance)
{
	// 20110418-fishy tweaking to avoid extensions!
	if (distance > 1000)
		return INFINITY;
	if (distance > 780) // 800
		return FIFTYPLUSFIFTY;
	if (distance > 580) // 500
		return FIFTYPLUSTHIRTY;
	if (distance > 480) // 500
		return FIFTYPLUSTEN;
	if (distance > 380) // 400
		return 500;
	if (distance > 280) // 300
		return THIRTYPLUSTEN;
	if (distance > 180) // 200
		return 300;
	if (distance > 100)
		return TENPLUSTEN;
	return 100;
}

unsigned find_slack(unsigned distance)
{
	switch (cable_select(distance)) {
	case FIFTYPLUSFIFTY:
		return 1000 - distance;
	case FIFTYPLUSTHIRTY:
		return 800 - distance;
	case FIFTYPLUSTEN:
		return 600 - distance;
	case 500:
		return 500 - distance;
	case THIRTYPLUSTEN:
		return 400 - distance;
	case 300:
		return 300 - distance;
	case TENPLUSTEN:
		return 200 - distance;
	case 100:
		return 100 - distance;
	}
}

unsigned find_cost(unsigned distance)
{
	//return find_slack(distance);	

#if TRUNCATE_METRIC
	return cable_select(distance);
#else
	return distance;
//	return ((distance + 90) / 100) * 100;
#endif
}

std::pair<unsigned, unsigned> find_optimal_cost(key &k)
{
	unsigned best_cost = 999999990, best_distance = 999999990;
	unsigned best_this_cost = 0, best_this_distance = 0;
	int best = -1;
	unsigned char db = k.distro_bound;
	unsigned char dbh = k.distro_bound_hard;
	
	if (k.sw_ind == switches.size())
		return std::make_pair(0, 0);
	if (cache.count(k)) {
		++cache_hits;
		return std::make_pair(cache[k].cost, cache[k].distance);
	}

	bool next_changes_row = (k.sw_ind + 1U < switches.size() && switches[k.sw_ind].row != switches[k.sw_ind + 1].row);
	
	for (unsigned char i = dbh; i < dbh + 2 && i < NUM_DISTRO; ++i) {
		if (k.ports_left[i] == 0)
			continue;
		
		unsigned this_distance = find_distance(switches[k.sw_ind], i);
		unsigned this_cost = find_cost(this_distance);
		
		--(k.ports_left[i]);
		++(k.sw_ind);
		
		k.distro_bound = std::max(i, db);
		if (next_changes_row) {
			k.distro_bound_hard = k.distro_bound;
		}

		std::pair<unsigned, unsigned> oc = find_optimal_cost(k);
		unsigned cost = this_cost + oc.first;
		unsigned distance = this_distance + oc.second;
		--(k.sw_ind);
		++(k.ports_left[i]);

		if (best == -1 || cost < best_cost || (cost == best_cost && distance < best_distance)) {
			best = i;
			best_cost = cost;
			best_distance = distance;
			best_this_cost = this_cost;
			best_this_distance = this_distance;
		}
	}
	k.distro_bound = db;
	k.distro_bound_hard = dbh;

	value v = { best, best_cost, best_distance, best_this_cost, best_this_distance };
	cache[k] = v;

	return std::make_pair(best_cost, best_distance);
}
			
std::string port_name(unsigned distro, unsigned portnum)
{
	char buf[16];
	//sprintf(buf, "d%u/p%u ", distro, portnum);
	//return buf;

	if (distro == 0) {
		int ports_per_card = 48;
		if (portnum <= ports_per_card) {
			sprintf(buf, "corengw g7/%u", portnum);
		} else if (portnum <= ports_per_card*2) {
			sprintf(buf, "corengw g8/%u", portnum - ports_per_card);
		} else {
			sprintf(buf, "corengw g9/%u", portnum - ports_per_card*2);
		}
	} else if (distro == 1) {
		int ports_per_card = 48;
		if (portnum <= ports_per_card) {
			sprintf(buf, "coresgw g4/%u", portnum);
		} else if (portnum <= ports_per_card*2) {
			sprintf(buf, "coresgw g7/%u", portnum - ports_per_card);
		} else if (portnum <= ports_per_card*3) {
			sprintf(buf, "coresgw g8/%u", portnum - ports_per_card*2);
		} else {
			sprintf(buf, "coresgw g9/%u", portnum - ports_per_card*3);
		}
	}

	return buf;
}

int main(int argc, char **argv)
{
	struct key start;
	FILE *patchlist = fopen("patchlist.txt", "w");
	FILE *switchlist = fopen("switches.txt", "w");
#if TRUNCATE_METRIC
	std::map<unsigned, unsigned> cable_count;
#endif
	unsigned total_slack = 0;

#if 1
	// used for placement optimization
	distro_placements[0] = atoi(argv[1]);
	distro_placements[1] = atoi(argv[2]);
#endif

	for (unsigned i = 1; i <= 42; ++i) {
		if (!(i >= 1 && i <= 5)) {
			switches.push_back(sw(i, 0));
			switches.push_back(sw(i, 1));
		}
		if (!(i >= 14 && i <= 21) &&
		    !(i >= 38)) {
		    //!(i == 43)) {
			switches.push_back(sw(i, 2));
			switches.push_back(sw(i, 3));
		}

		/*
		if (!(i >= 37 )) {
			switches.push_back(sw(i, 0));
			switches.push_back(sw(i, 1));
		}
		*/
	}

	start.sw_ind = 0;
	start.distro_bound = 0;
	start.distro_bound_hard = 0;
	start.ports_left[0] = 72;
	start.ports_left[1] = 96;

#if 0
	// split distro
	start.ports_left[0] = 22;
	start.ports_left[1] = 22;
	start.ports_left[2] = 42;
	start.ports_left[3] = 37;
	start.ports_left[4] = 42;
	start.ports_left[5] = 22;
	start.ports_left[6] = 22; 
#endif
	
	printf("Finding optimal layout for %u switches\n", switches.size());
	find_optimal_cost(start);
	printf("%u cache nodes, %u cache hits.\n", cache.size(), cache_hits);

	key k = start;
	int last_row = 0, last_num = -1;
	int management_address = 2;
	for (unsigned i = 0; i < switches.size(); ++i) {
		value v = cache[k];
		bool next_changes_row = (k.sw_ind + 1U < switches.size() && switches[k.sw_ind].row != switches[k.sw_ind + 1].row);

		if (cache.count(k) == 0) {
			fprintf(stderr, "FATAL: no solutions!\n");
			exit(1);
		}
		
		k.distro_bound = std::max(k.distro_bound, v.distro);
		if (next_changes_row)
			k.distro_bound_hard = k.distro_bound;
		
		if (last_row != switches[k.sw_ind].row) {
			if (last_row == 13 || last_row == 21 || last_row == 29 ) {
				printf("\n");
			}
			printf("\n%2u (%2u-%2u)    ", switches[k.sw_ind].row, switches[k.sw_ind].row * 2 - 1, switches[k.sw_ind].row * 2 + 0 );
			last_num = -1;
		}
		for (int j = last_num; j + 1 < switches[k.sw_ind].num; ++j) {
			printf("%19s", "");
		}

		char distro_marker = ' ';
		if (switches[k.sw_ind].num == 1) {
			for (int d = 0; d < NUM_DISTRO; ++d) {
				if (switches[k.sw_ind].row == distro_placements[d]) {
					distro_marker = '*';
				}
			}
		}

		
		printf("[%um%u ", v.distro + 33, v.distro);
#if TRUNCATE_METRIC
		if (cable_select(v.this_distance) == INFINITY)
			printf("(XXXXX) ");
		else if (cable_select(v.this_distance) == FIFTYPLUSFIFTY)
			printf("(50+50) ");
		else if (cable_select(v.this_distance) == FIFTYPLUSTHIRTY)
			printf("(50+30) ");
		else if (cable_select(v.this_distance) == FIFTYPLUSTEN)
			printf("(50+10) ");
		else if (cable_select(v.this_distance) == THIRTYPLUSTEN)
			printf("(30+10) ");
		else if (cable_select(v.this_distance) == TENPLUSTEN)
			printf("(10+10) ");
		else
			printf("(%-3u  ) ", cable_select(v.this_distance) / 10);
		printf("(%3.1f)%c  ", v.this_distance / 10.0, distro_marker);
		total_slack += find_slack(v.this_distance);
		cable_count[cable_select(v.this_distance)]++;
#else
		total_slack += find_slack(v.this_distance);
		printf("(%3.1f)%c  ", v.this_cost / 10.0, distro_marker);
#endif
				
		last_row = switches[k.sw_ind].row;
		last_num = switches[k.sw_ind].num;
			
		++(k.sw_ind);
		--k.ports_left[v.distro];

		int port_num = ++num_ports_used[v.distro];
		if (port_num <= 24) {
			fprintf(patchlist, "e%u-%u %s %s\n",
			        last_row * 2 - 1, last_num + 1,
				port_name(v.distro, port_num).c_str(),
				port_name(v.distro, port_num + 48).c_str());
		} else if (port_num <= 48) {
			fprintf(patchlist, "e%u-%u %s %s\n",
			        last_row * 2 - 1, last_num + 1,
				port_name(v.distro, port_num).c_str(),
				port_name(v.distro, port_num + 96 - 24).c_str());
		} else {
			fprintf(patchlist, "e%u-%u %s %s\n",
			        last_row * 2 - 1, last_num + 1,
				port_name(v.distro, port_num + 24).c_str(),
				port_name(v.distro, port_num + 96 - 24).c_str());
		}

		switch (last_num + 1) {
		case 1:
			fprintf(switchlist, "188.0.%u.0 25 e%u-%u 188.0.101.%u\n",
	 			last_row * 2 - 1, last_row * 2 - 1, last_num + 1, management_address++);
			break;
		case 2:
			fprintf(switchlist, "188.0.%u.128 25 e%u-%u 188.0.101.%u\n",
	 			last_row * 2 - 1, last_row * 2 - 1, last_num + 1, management_address++);
			break;
		case 3:
			fprintf(switchlist, "188.0.%u.0 25 e%u-%u 188.0.101.%u\n",
	 			last_row * 2, last_row * 2 - 1, last_num + 1, management_address++);
			break;
		case 4:
			fprintf(switchlist, "188.0.%u.128 25 e%u-%u 188.0.101.%u\n",
	 			last_row * 2, last_row * 2 - 1, last_num + 1, management_address++);
			break;
		}
	}
	printf("\n");
	printf("[%um\n", 37);

#if TRUNCATE_METRIC
	cable_count[100] += cable_count[TENPLUSTEN];
	cable_count[100] += cable_count[TENPLUSTEN];

	cable_count[100] += cable_count[THIRTYPLUSTEN];
	cable_count[300] += cable_count[THIRTYPLUSTEN];

	cable_count[100] += cable_count[FIFTYPLUSTEN];
	cable_count[500] += cable_count[FIFTYPLUSTEN];
	
	cable_count[300] += cable_count[FIFTYPLUSTHIRTY];
	cable_count[500] += cable_count[FIFTYPLUSTHIRTY];
	
	cable_count[500] += cable_count[FIFTYPLUSFIFTY] * 2;

	printf("\n");
	printf("10m: %3u\n", cable_count[100]);
	printf("30m: %3u\n", cable_count[300]);
	printf("50m: %3u\n", cable_count[500]);
	printf("Extensions: %u\n", cable_count[TENPLUSTEN] + cable_count[THIRTYPLUSTEN] + cable_count[FIFTYPLUSTEN]);
	printf("\n");

	if (cable_count[INFINITY] > 0) {
		printf("Total cost: Impossible\n");
	} else {
		unsigned total_cable = 100 * cable_count[100] + 300 * cable_count[300] + 500 * cable_count[500];
		printf("Total cost: %.1fm\n", 10.0 * cable_count[100] + 30.0 * cable_count[300] + 50.0 * cable_count[500] + 5.0 * (cable_count[TENPLUSTEN] + cable_count[THIRTYPLUSTEN] + cable_count[FIFTYPLUSTEN]));
		printf("Total cable: %.1fm\n", total_cable / 10.0);
		printf("Total slack: %.1fm (%.2f%%)\n", total_slack / 10.0, 100.0 * double(total_slack) / double(total_cable));
	}
	printf ("Remaining ports on distro #1: %d\n", k.ports_left[0]);
	printf ("Remaining ports on distro #2: %d\n", k.ports_left[1]);

#endif
}
